/*
 * Rate converter plugin using Public Parrot Hack
 *
 * Copyright (c) 2006 by Takashi Iwai <tiwai@suse.de>
 * Copyright (c) 2007 by Jean-Marc Valin <jean-marc.valin@usherbrooke.ca>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <samplerate.h>
#include <alsa/asoundlib.h>
#include <alsa/pcm_rate.h>

#include "speex_resampler.h"

struct rate_src {
	int quality;
	unsigned int channels;
        SpeexResamplerState *st;
};

static snd_pcm_uframes_t input_frames(void *obj, snd_pcm_uframes_t frames)
{
   int num, den;
   struct rate_src *rate = obj;
   if (frames == 0)
      return 0;
   speex_resampler_get_ratio(rate->st, &num, &den);
   return (snd_pcm_uframes_t)(frames*num/den);
   //return (snd_pcm_uframes_t)(frames*den/num);
}

static snd_pcm_uframes_t output_frames(void *obj, snd_pcm_uframes_t frames)
{
   int num, den;
   struct rate_src *rate = obj;
   if (frames == 0)
      return 0;
   speex_resampler_get_ratio(rate->st, &num, &den);
   return (snd_pcm_uframes_t)(frames*den/num);
   //return (snd_pcm_uframes_t)(frames*num/den);
}

static void pcm_src_free(void *obj)
{
   struct rate_src *rate = obj;
   if (rate->st)
   {
      speex_resampler_destroy(rate->st);
      rate->st = NULL;
   }
}

static int pcm_src_init(void *obj, snd_pcm_rate_info_t *info)
{
   struct rate_src *rate = obj;
   
   if (! rate->st || rate->channels != info->channels) {
      if (rate->st)
         speex_resampler_destroy(rate->st);
      rate->channels = info->channels;
      rate->st = speex_resampler_init_frac(rate->channels, info->in.period_size, info->out.period_size, info->in.rate, info->out.rate, rate->quality);
      if (! rate->st)
         return -EINVAL;
   }

   return 0;
}

static int pcm_src_adjust_pitch(void *obj, snd_pcm_rate_info_t *info)
{
   struct rate_src *rate = obj;
   speex_resampler_set_rate_frac(rate->st, info->in.period_size, info->out.period_size, info->in.rate, info->out.rate);
   return 0;
}

static void pcm_src_reset(void *obj)
{
   struct rate_src *rate = obj;
   speex_resampler_reset_mem(rate->st);
}

static void pcm_src_convert_s16(void *obj, int16_t *dst, unsigned int dst_frames,
				const int16_t *src, unsigned int src_frames)
{
   struct rate_src *rate = obj;
   speex_resampler_process_interleaved_int(rate->st, src, &src_frames, dst, &dst_frames);
}

static void pcm_src_close(void *obj)
{
   free(obj);
}

static snd_pcm_rate_ops_t pcm_src_ops = {
	.close = pcm_src_close,
	.init = pcm_src_init,
	.free = pcm_src_free,
	.reset = pcm_src_reset,
	.adjust_pitch = pcm_src_adjust_pitch,
	.convert_s16 = pcm_src_convert_s16,
	.input_frames = input_frames,
	.output_frames = output_frames,
};

static int pcm_src_open(unsigned int version, void **objp,
			snd_pcm_rate_ops_t *ops, int quality)
{
	struct rate_src *rate;

	if (version != SND_PCM_RATE_PLUGIN_VERSION) {
		fprintf(stderr, "Invalid rate plugin version %x\n", version);
		return -EINVAL;
	}

	rate = calloc(1, sizeof(*rate));
	if (! rate)
		return -ENOMEM;
	rate->quality = quality;

	*objp = rate;
	*ops = pcm_src_ops;
	return 0;
}

int SND_PCM_RATE_PLUGIN_ENTRY(speexrate) (unsigned int version, void **objp,
					   snd_pcm_rate_ops_t *ops)
{
	return pcm_src_open(version, objp, ops, 3);
}

int SND_PCM_RATE_PLUGIN_ENTRY(speexrate_best) (unsigned int version, void **objp,
						snd_pcm_rate_ops_t *ops)
{
	return pcm_src_open(version, objp, ops, 10);
}

int SND_PCM_RATE_PLUGIN_ENTRY(speexrate_medium) (unsigned int version, void **objp,
						  snd_pcm_rate_ops_t *ops)
{
	return pcm_src_open(version, objp, ops, 5);
}
